/***************************************************************************
 *   Copyright (C) 2007 by David Lopez Vinacua                             *
 *   david@bajavista.com                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/* * * * * * * * * * * * * * * * * * * * * *
 *  vfsrecycle_clean.c                     *
 *  Code and Data for recycle can cleaning *
 * * * * * * * * * * * * * * * * * * * * * *
 *
 * Created:
 *  130307 David Lopez
 * Modified:
 *
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include <time.h>
#include <ctype.h>

#include "vfsrecycle_clean.h"


void Usage(char *apName)
{
  fprintf(stdout,"Usage:\n");
  fprintf(stdout,"  %s [options]\n",apName);
  fprintf(stdout,"Options:\n");
  fprintf(stdout," -can=<trashcan name>\n");
  fprintf(stdout," -g=<grace period>\n");
  fprintf(stdout," -r=<Storage root>\n");
  fprintf(stdout," -a=<Attributes file preffix>\n");
  fprintf(stdout," -c=<Storage class to scan> (may be several ones)\n");
}

short ReadParams(int argc, char *argv[], t_vfsr_par *par)
/* Reads command line params into par */
{
  int               cp;
  char             *c;
  short             res;
  int               i;
  
  // First set defaults
  strcpy(&par->tcN[0],"recycle");
  par->sR[0]='\0';
  par->sC[0]='\0';
  par->scp=0;
  par->gD=3;
  // Now override with command line specific values
  res=0;
  cp=1;
  while ( (cp<argc) && (res==0) )
    {
    c=argv[cp];
    if (*c!='-')
       {
       res=1;
       break;
       }
    c++;
    switch (*c)
      {
      case 'a': // Attributes file prefix
        if (*(c+1)=='=')
           {
           c+=2;
           strncpy(&par->aP[0],c,sizeof(par->aP)-1);
           par->aP[sizeof(par->aP)-1]='\0';
           }
        break;
      case 'c':
        if (strncmp(c,"can=",4)==0)
           {    // Trash can name
           c+=4;
           strncpy(&par->tcN[0],c,sizeof(par->tcN)-1);
           par->tcN[sizeof(par->tcN)-1]='\0';
           }
        else if (*(c+1)=='=')
           {    // Storage class
           c+=2;
           i=strlen(c);
           if (par->scp+i+1>=sizeof(par->sC))
              {
              fprintf(stderr,"Storage class buffer overflow (too many or too long)\n\n");
              res=2;
              }
           else
              {
              strcpy(&par->sC[par->scp],c);
              par->scp+=strlen(c)+1;
              }
           }
        else
           res=1;
        break;
      case 'g':
        c++;
        if (*c=='=')
           {
           c++;
           i=0;
           if (sscanf(c,"%u",&i)!=1)
              res=1;
           else
              par->gD=i;
           }
        else
           res=1;
        break;
      case 'r':
        c++;
        if (*c=='=')
           {
           c++;
           strncpy(&par->sR[0],c,sizeof(par->sR)-2);
           par->sR[sizeof(par->sR)-2]='\0';
           if (par->sR[0])
              {
              c=strchr(&par->sR[0],'\0');
              if (*(c-1)!='/')
                 {
                 *c='/';
                 *(c+1)='\0';
                 }
              }
           }
        else
           res=1;
        break;
      default:
        res=1;
      }
    if (!res)
       cp++;
    }
  switch (res)
    {
    case 1:
      fprintf(stderr,"Bad param (%u: %s)\n\n",cp,argv[cp]);
      Usage(argv[0]);
      break;
    }
  return res;
}

int GetNum(char **s)
/* Extracts a number from c, returns -1 on failure. c points just after the number */
{
  char             *c;
  int               i;
  
  if (!s || !(*s))
     return -1;
  c=*s;
  while (*c && ((*c<'0') || (*c>'9')))
    c++;
  if (!(*c))
     return -1;
  i=0;
  while (*c && (*c>='0') && (*c<='9'))
     {
     i=10*i+(*c-'0');
     c++;
     }
  return i;
}

short ReadAttributeFile(char *fn, t_vfs_atr *a)
/* Tries to open and parse attributes file. Returns 0 on success */
{
  FILE             *f;
  char             *c,*d,*e,*g;
  char              s[256];
  int               i;
  
  f=fopen(fn,"r");
  if (f)
     {
     while ( (fgets(&s[0],sizeof(s),f)) )
       {
       c=strchr(s,'#');
       if (c)
          *c='\0';
       c=strchr(s,'\n');
       if (c)
          *c='\0';
       c=&s[0];
       while (*c && ((*c==' ') || (*c=='\t')))
         c++;
       if (*c)
          {
          d=strchr(c,'=');
          if (d)
             {
             *d='\0';
             e=c;
             g=e;
             while (e<d)
               {
               if ( (*e!=' ') && (*e!='\t') )
                  {
                  *g=tolower(*e);
                  g++;
                  }
               e++;
               }
             *g='\0';
             if (strcmp(c,"gracedays")==0)
                {
                c=d+1;
                while (*c && ((*c==' ') || (*c=='\t')))
                  c++;
                i=GetNum(&c);
                if (i!=-1)
                   a->gD=i;
                }
             }
          }
       }
     fclose(f);
     return 0;
     }
  return -1;
}

short Scan(t_vfsr_par *par)
/* Scans every share and removes deleted files older than grace time */
{
  short             cp;
  char              fn[256];
  DIR              *d,*d2;
  struct dirent    *de;
  t_vfs_atr         atr;
  struct stat       st;
  short             fn1,l,max,tcl,fn2;
  char              rm;
  time_t            rt,dt;
  
  if (!par->scp)
     {
     fprintf(stderr,"No storage classes to scan\n");
     return 1;
     }
  dt=time(NULL);
  cp=0;
  max=sizeof(fn)-1;
  tcl=1+strlen(&par->tcN[0]);   // Length of trashcan name, including preceding slash or null terminator
  while (cp<par->scp)           // For each storage class, get all its shares, and then clean every trash can inside
    {
    // First try to parse attribute file for current class
    sprintf(&fn[0],"%s%s%s",par->sR,&par->aP[0],&par->sC[cp]);
    // Set defaults first
    bzero(&atr,sizeof(atr));
    atr.gD=par->gD;
    ReadAttributeFile(&fn[0],&atr); // Alter defaults from specific attributes file
    rt=dt-24*3600*atr.gD;    // Grace time limit
    // Now go to work for that class
    sprintf(&fn[0],"%s%s",par->sR,&par->sC[cp]);
    cp+=strlen(&par->sC[cp])+1;
    d=opendir(&fn[0]);
    if (d)
       {
       fn1=strlen(&fn[0]);
       fn[fn1]='/';
       fn1++;
       while ( (de=readdir(d)) )
         {
         if (de->d_name[0]!='.')
            {
            l=strlen(&de->d_name[0]);
            if (fn1+l+tcl+1<max)
               {
               memcpy(&fn[fn1],&de->d_name[0],l);
               fn[fn1+l]='/';
               memcpy(&fn[fn1+l+1],&par->tcN[0],tcl);   // Includes also null terminator (fn2 is one more than tcN length)
               // Now we scan this trash can for files to remove
               d2=opendir(&fn[0]);
               if (d2)
                  {
                  fn2=fn1+l+tcl;
                  fn[fn2]='/';
                  fn2++;
                  rm=1;
                  while (rm)
                    {
                    rm=0;
                    while ( (de=readdir(d2)) )
                       {
                       if (de->d_name[0]!='.')
                          {
                          l=strlen(&de->d_name[0]);
                          if (fn2+l+1<max)
                             {
                             memcpy(&fn[fn2],&de->d_name[0],l+1);   // Include terminator
                             if (stat(&fn[0],&st)==0)
                                {
                                if ( (S_ISREG(st.st_mode)) && (/*st.st_atime*/st.st_mtime<rt) )
                                   {
                                   fprintf(stdout," [Deleting] %s\n",&fn[0]);
                                   unlink(&fn[0]);
                                   rm=1;
                                   }
                                }
                             }
                          }
                       }
                    if (rm)     // Start from the beginning (rm changes directory if it is not cached)
                       rewinddir(d2);
                    }
                  closedir(d2);
                  }
               }
            }
         }
       closedir(d);
       }
    }
  return 0;
}

int main(int argc, char *argv[])
{
  t_vfsr_par        par;
  
  if (argc<2)
     {
     Usage(argv[0]);
     return 0;
     }
  bzero(&par,sizeof(par));
  if (ReadParams(argc,argv,&par))
     return 1;
  Scan(&par);
  return 0;
}
